/**********************************************************************
 * pgchem_setup.sql unified type, operator and function setup file for pgchem GiST
 *
 * Copyright (c) 2004,2009 by Ernst-G. Schmid
 *
 * This file is part of the pgchem::tigress project.
 * For more information, see
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 ************************************************************************/

DROP TYPE molecule CASCADE;
DROP TYPE molfp CASCADE;

CREATE TYPE molecule;
CREATE TYPE molfp;

CREATE FUNCTION molfp_in(cstring)
    RETURNS molfp
    AS 'libpgchem'
    LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION molfp_out(molfp)
    RETURNS cstring
   AS 'libpgchem'
    LANGUAGE C IMMUTABLE STRICT;

CREATE TYPE molfp (
   input = molfp_in,
   output = molfp_out,
   internallength = 192,
   storage = PLAIN
   );

CREATE FUNCTION molecule_in(cstring)
    RETURNS molecule
    AS 'libpgchem'
    LANGUAGE C IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION molecule_out(molecule)
    RETURNS cstring
   AS 'libpgchem'
    LANGUAGE C IMMUTABLE STRICT;

CREATE FUNCTION molecule_recv(internal)
   RETURNS molecule
  AS 'libpgchem'
   LANGUAGE C IMMUTABLE STRICT;

CREATE FUNCTION molecule_send(molecule)
   RETURNS bytea
  AS 'libpgchem'
   LANGUAGE C IMMUTABLE STRICT;

CREATE TYPE molecule (
   input = molecule_in,
   output = molecule_out,
   --receive = molecule_recv,
   --send = molecule_send,
   internallength = VARIABLE,
   storage = EXTENDED
   );

CREATE FUNCTION molfp_compress(internal)
RETURNS internal
AS 'libpgchem'
LANGUAGE 'C';

CREATE FUNCTION molfp_decompress(internal)
RETURNS internal
AS 'libpgchem'
LANGUAGE 'C';

CREATE FUNCTION molfp_penalty(internal,internal,internal)
RETURNS internal
AS 'libpgchem'
LANGUAGE 'C' WITH (isstrict);

CREATE FUNCTION molfp_picksplit(internal,internal)
RETURNS internal
AS 'libpgchem'
LANGUAGE 'C' WITH (isstrict);

CREATE FUNCTION molfp_union(internal, internal)
RETURNS internal
AS 'libpgchem'
LANGUAGE 'C';

CREATE FUNCTION molfp_same(internal, internal, internal)
RETURNS internal
AS 'libpgchem'
LANGUAGE 'C';

CREATE FUNCTION molfp_consistent(internal,internal,int4)
RETURNS bool
AS 'libpgchem'
LANGUAGE 'C';

CREATE FUNCTION molecule_contained_in(molecule,molecule)
RETURNS bool
AS 'libpgchem'
LANGUAGE 'C' with (isstrict);

CREATE FUNCTION molecule_contains(molecule,molecule)
RETURNS bool
AS 'libpgchem'
LANGUAGE 'C' with (isstrict);

CREATE FUNCTION molecule_equals(molecule,molecule)
RETURNS bool
AS 'libpgchem'
LANGUAGE 'C' with (isstrict);

CREATE FUNCTION molecule_similarity(molecule,molecule)
RETURNS double precision
AS 'libpgchem'
LANGUAGE 'C' with (isstrict);

CREATE OPERATOR <= (
		 LEFTARG = molecule,
		 RIGHTARG = molecule,
		 PROCEDURE = molecule_contained_in,
                 COMMUTATOR = >=,
		 RESTRICT = contsel,
		 JOIN = contjoinsel
);

CREATE OPERATOR >= (
		 LEFTARG = molecule,
		 RIGHTARG = molecule,
		 PROCEDURE = molecule_contains,
                 COMMUTATOR = <=,
		 RESTRICT = contsel,
		 JOIN = contjoinsel
);

CREATE OPERATOR = (
		 LEFTARG = molecule,
		 RIGHTARG = molecule,
		 PROCEDURE = molecule_equals,
		 COMMUTATOR = =,
		 RESTRICT = eqsel,
		 JOIN = eqjoinsel
);

CREATE OPERATOR @ (
    leftarg = molecule,
    rightarg = molecule,
    procedure = molecule_similarity
);

CREATE OPERATOR CLASS gist_molecule_ops
DEFAULT FOR TYPE molecule USING gist
AS
        OPERATOR        6       =        RECHECK,
        OPERATOR        7       >=       RECHECK,
        OPERATOR        8       <=       RECHECK,
        FUNCTION        1       molfp_consistent (internal, internal, int4),
        FUNCTION        2       molfp_union (internal, internal),
        FUNCTION        3       molfp_compress (internal),
        FUNCTION        4       molfp_decompress (internal),
        FUNCTION        5       molfp_penalty (internal, internal, internal),
        FUNCTION        6       molfp_picksplit (internal, internal),
        FUNCTION        7       molfp_same (internal, internal, internal),
        STORAGE		molfp;

CREATE OR REPLACE FUNCTION molecule_in(text)
  RETURNS molecule AS
'libpgchem', 'molecule_in_text'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION molecule_in(character varying)
  RETURNS molecule AS
'libpgchem', 'molecule_in_varchar'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION molecule_in(bytea)
  RETURNS molecule AS
'libpgchem', 'molecule_in_bytea'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE CAST (text AS molecule)
  WITH FUNCTION molecule_in(text);

CREATE CAST (character varying AS molecule)
  WITH FUNCTION molecule_in(character varying);

CREATE CAST (bytea AS molecule)
  WITH FUNCTION molecule_in(bytea);

CREATE OR REPLACE FUNCTION add_hydrogens(molecule, boolean, boolean)
  RETURNS molecule AS
'libpgchem', 'pgchem_add_hydrogens'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION exactmass(molecule)
  RETURNS double precision AS
'libpgchem', 'pgchem_exactmass'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION fgroup_codes(molecule)
  RETURNS text AS
'libpgchem', 'pgchem_fgroup_codes_a'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION is_2d(molecule)
  RETURNS boolean AS
'libpgchem', 'pgchem_2D'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION is_3d(molecule)
  RETURNS boolean AS
'libpgchem', 'pgchem_3D'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION is_chiral(molecule)
  RETURNS boolean AS
'libpgchem', 'pgchem_is_chiral'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION is_nostruct(molecule)
  RETURNS boolean AS
'libpgchem', 'pgchem_is_nostruct'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION inchi(molecule)
  RETURNS text AS
'libpgchem', 'pgchem_molecule_to_inchi'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION canonical_smiles(molecule, boolean)
  RETURNS text AS
'libpgchem', 'pgchem_molecule_to_canonical_smiles'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION smiles(molecule)
  RETURNS text AS
'libpgchem', 'pgchem_molecule_to_smiles'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION v3000(molecule)
  RETURNS text AS
'libpgchem', 'pgchem_molecule_to_V3000'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION molfile(molecule)
  RETURNS text AS
'libpgchem', 'pgchem_molecule_to_molfile'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION molformula(molecule)
  RETURNS text AS
'libpgchem', 'pgchem_hillformula'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION molweight(molecule)
  RETURNS double precision AS
'libpgchem', 'pgchem_molweight'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION molkeys_long(molecule, boolean, boolean, boolean)
  RETURNS text AS
'libpgchem', 'pgchem_ms_fingerprint_long_a'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION molkeys_long(molecule)
  RETURNS text AS
$BODY$
BEGIN
RETURN molkeys_long($1,false,false,false);
END;
$BODY$
  LANGUAGE 'plpgsql' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION molkeys_short(molecule)
  RETURNS text AS
'libpgchem', 'pgchem_ms_fingerprint_short_a'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION number_of_atoms(molecule)
  RETURNS integer AS
'libpgchem', 'pgchem_num_atoms'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION number_of_bonds(molecule)
  RETURNS integer AS
'libpgchem', 'pgchem_num_bonds'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION number_of_heavy_atoms(molecule)
  RETURNS integer AS
'libpgchem', 'pgchem_num_heavy_atoms'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION number_of_rotatable_bonds(molecule)
  RETURNS integer AS
'libpgchem', 'pgchem_num_rotatable_bonds'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION pgchem_barsoi_version()
  RETURNS cstring AS
'libpgchem', 'pgchem_barsoi_version'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION pgchem_version()
  RETURNS cstring AS
'libpgchem', 'pgchem_version'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION remove_hydrogens(molecule, boolean)
  RETURNS molecule AS
'libpgchem', 'pgchem_remove_hydrogens'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION strip_salts(molecule, boolean)
  RETURNS molecule AS
'libpgchem', 'pgchem_strip_salts'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION total_charge(molecule)
  RETURNS integer AS
'libpgchem', 'pgchem_total_charge'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION validate_cas_no(character varying)
  RETURNS boolean AS
$BODY$
DECLARE checksum_from_cas_no varchar;
DECLARE cas_no_left varchar;
DECLARE cas_no_right varchar;
DECLARE cas_no_full varchar;
DECLARE tmpsum int;
DECLARE position_multiplier int;
DECLARE caslen int;
BEGIN
caslen:=length($1);

IF caslen<5 OR caslen>12 THEN RETURN FALSE;
END IF;

checksum_from_cas_no:=split_part($1,'-',3)::int;
cas_no_left:=split_part($1,'-',1);
cas_no_right:=split_part($1,'-',2);
cas_no_full:=cas_no_left || cas_no_right;

if(length(cas_no_left)>7 OR length(cas_no_right)>2 OR length(checksum_from_cas_no)!=1) THEN 
return false; 
END IF;

caslen:=length(cas_no_full);
tmpsum:=0;
position_multiplier:=1;

 FOR i IN REVERSE caslen..1 LOOP
  tmpsum:=tmpsum+substr(cas_no_full,i,1)::int*position_multiplier;
  position_multiplier:=position_multiplier+1;
 END LOOP;
 RETURN tmpsum % 10 = checksum_from_cas_no::int;
END;
$BODY$
  LANGUAGE 'plpgsql' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION logP(molecule)
  RETURNS double precision AS
'libpgchem', 'pgchem_logP'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION MR(molecule)
  RETURNS double precision AS
'libpgchem', 'pgchem_MR'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION PSA(molecule)
  RETURNS double precision AS
'libpgchem', 'pgchem_PSA'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION migrate_molecule(bytea)
  RETURNS molecule AS
'libpgchem','pgchem_molecule_to_new_molecule'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION SMARTSmatch(text, molecule)
  RETURNS boolean AS
'libpgchem', 'pgchem_smartsfilter'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION SMARTSmatch_count(text, molecule)
  RETURNS integer AS
'libpgchem', 'pgchem_smartsfilter_count'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION molhash(molecule)
  RETURNS text AS
'libpgchem', 'pgchem_molhash'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION fpstring(molecule)
  RETURNS bit varying AS
'libpgchem', 'pgchem_fp_out'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION number_of_h_acceptors(molecule)
  RETURNS integer AS
'libpgchem', 'pgchem_num_H_acceptors'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION number_of_h_donors(molecule)
  RETURNS integer AS
'libpgchem', 'pgchem_num_H_donors'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION lipinsky(mol molecule)
  RETURNS text AS
$BODY$
DECLARE parameters text;
BEGIN
parameters := '';

IF number_of_H_donors(mol) > 5 THEN
parameters := parameters || 'A';
END IF;
IF molweight(mol) > 500 THEN
parameters := parameters || 'B';
END IF;
IF logP(mol) > 5.0 THEN
parameters := parameters || 'C';
END IF;
IF number_of_H_acceptors(mol) > 10 THEN
parameters := parameters || 'D';
END IF;

RETURN parameters;
END;
$BODY$
  LANGUAGE 'plpgsql' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION mutabor(molecule)
  RETURNS molecule AS
'libpgchem', 'pgchem_mutate_fp'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION delebor(molecule)
  RETURNS molecule AS
'libpgchem', 'pgchem_blank_fp'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION fp2string(struct molecule)
  RETURNS bit varying AS
$BODY$
DECLARE fp bit varying;
BEGIN
fp := substring(fpstring(struct) for 1024);
RETURN fp;
END;
$BODY$
  LANGUAGE 'plpgsql' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION fp3string(struct molecule)
  RETURNS bit varying AS
$BODY$
DECLARE fp bit varying;
BEGIN
fp := substring(fpstring(struct) from 1025);
RETURN fp;
END;
$BODY$
  LANGUAGE 'plpgsql' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION nbits_set(bit varying)
  RETURNS integer AS
'libpgchem', 'pgchem_nbits_set'
  LANGUAGE 'c' IMMUTABLE STRICT;

CREATE OR REPLACE FUNCTION disconnected(molecule)
  RETURNS boolean AS
'libpgchem', 'pgchem_disconnected'
  LANGUAGE 'c' IMMUTABLE STRICT;




